def get_params():
    params = []
    # your indicator
    video_type = 1
    #------- fill your params ------

    # ------------------------------ 
    params.append(video_type)
    return params 

